(function (global, factory) {
    typeof exports === 'object' && typeof module !== 'undefined' ? factory(exports, require('rxjs'), require('rxjs/operators'), require('@rx-angular/cdk/coercing')) :
    typeof define === 'function' && define.amd ? define('@rx-angular/cdk/notifications', ['exports', 'rxjs', 'rxjs/operators', '@rx-angular/cdk/coercing'], factory) :
    (global = typeof globalThis !== 'undefined' ? globalThis : global || self, factory((global['rx-angular'] = global['rx-angular'] || {}, global['rx-angular']['cdk-notifications'] = {}), global.rxjs, global.rxjs.operators, global['rx-angular']['cdk-coercing']));
}(this, (function (exports, rxjs, operators, coercing) { 'use strict';

    function toRxErrorNotification(error, value) {
        return {
            value: value,
            kind: "error" /* Error */,
            hasValue: !!value || false,
            complete: false,
            error: error || true,
        };
    }
    function toRxSuspenseNotification(value) {
        return {
            value: value,
            kind: "suspense" /* Suspense */,
            hasValue: !!value || false,
            complete: false,
            error: false,
        };
    }
    function toRxCompleteNotification(value) {
        return {
            value: value,
            kind: "complete" /* Complete */,
            hasValue: !!value || false,
            complete: true,
            error: false,
        };
    }

    /**
     * @internal
     *
     * A factory function returning an object to handle the process of switching templates by Notification channel.
     * You can next a Observable of `RxNotification` multiple times and merge them into the Observable exposed under `trigger$`
     *
     */
    function templateTriggerHandling() {
        var hotFlattened = coercing.coerceAllFactory(function () { return new rxjs.Subject(); }, operators.mergeAll());
        return {
            next: function (templateName) {
                hotFlattened.next(templateName);
            },
            trigger$: hotFlattened.values$.pipe(operators.share()),
        };
    }

    function rxMaterialize() {
        return function (o$) { return o$.pipe(operators.materialize(), operators.tap(function (_a) {
            var kind = _a.kind, error = _a.error;
            if (kind === 'E') {
                console.error(error);
            }
        }), operators.map(function (_a) {
            var value = _a.value, hasValue = _a.hasValue, error = _a.error, kind = _a.kind;
            var rxNotificationKind = notificationKindToRxNotificationKind(kind);
            return {
                value: value,
                hasValue: hasValue,
                error: error,
                kind: rxNotificationKind,
                complete: rxNotificationKind === "complete" /* Complete */,
            };
        })); };
    }
    function notificationKindToRxNotificationKind(kind) {
        switch (kind) {
            case 'C':
                return "complete" /* Complete */;
            case 'E':
                return "error" /* Error */;
            case 'N':
            default:
                return "next" /* Next */;
        }
    }

    /**
     * @internal
     */
    function createTemplateNotifier(withSuspense) {
        var observablesSubject = new rxjs.ReplaySubject(1);
        var firstRun = true;
        var latestNextValue;
        var wSuspense = function () { return withSuspense && withSuspense(); };
        var values$ = observablesSubject.pipe(operators.distinctUntilChanged(), 
        // handle null | undefined assignment and new Observable reset
        operators.map(function (observable$) {
            var isObs = rxjs.isObservable(observable$);
            // only pass through initial undefined value to filter out
            if (firstRun && (observable$ === undefined || observable$ === rxjs.NEVER)) {
                return wSuspense() ? rxjs.NEVER.pipe(operators.startWith(undefined)) : undefined;
            }
            var isNull = observable$ == null;
            if (isNull) {
                return rxjs.NEVER.pipe(operators.startWith(observable$));
            }
            return (isObs ? observable$ : rxjs.from(observable$)).pipe(function (o$) {
                if (!firstRun || (withSuspense && withSuspense())) {
                    return o$.pipe(operators.startWith(undefined));
                }
                return o$;
            });
        }), operators.tap(function () { return (firstRun = false); }), operators.filter(function (v) { return v !== undefined; }), operators.switchMap(function (o) {
            return o.pipe(operators.distinctUntilChanged(), rxMaterialize(), operators.map(function (notification) {
                latestNextValue =
                    notification.kind === "next" /* Next */
                        ? notification.value
                        : latestNextValue;
                if (notification.kind === "next" /* Next */ &&
                    latestNextValue === undefined) {
                    return toRxSuspenseNotification(latestNextValue);
                }
                return Object.assign(Object.assign({}, notification), { value: latestNextValue });
            }));
        }));
        return {
            next: function (observable) {
                observablesSubject.next(observable);
            },
            values$: values$,
        };
    }

    /**
     * Generated bundle index. Do not edit.
     */

    exports.createTemplateNotifier = createTemplateNotifier;
    exports.rxMaterialize = rxMaterialize;
    exports.templateTriggerHandling = templateTriggerHandling;
    exports.toRxCompleteNotification = toRxCompleteNotification;
    exports.toRxErrorNotification = toRxErrorNotification;
    exports.toRxSuspenseNotification = toRxSuspenseNotification;

    Object.defineProperty(exports, '__esModule', { value: true });

})));
//# sourceMappingURL=cdk-notifications.umd.js.map
