import { from, isObservable, NEVER, ReplaySubject, } from 'rxjs';
import { distinctUntilChanged, filter, map, startWith, switchMap, tap, } from 'rxjs/operators';
import { rxMaterialize } from './rx-materialize';
import { toRxSuspenseNotification } from './notification-transforms';
/**
 * @internal
 */
export function createTemplateNotifier(withSuspense) {
    const observablesSubject = new ReplaySubject(1);
    let firstRun = true;
    let latestNextValue;
    const wSuspense = () => withSuspense && withSuspense();
    const values$ = observablesSubject.pipe(distinctUntilChanged(), 
    // handle null | undefined assignment and new Observable reset
    map((observable$) => {
        const isObs = isObservable(observable$);
        // only pass through initial undefined value to filter out
        if (firstRun && (observable$ === undefined || observable$ === NEVER)) {
            return wSuspense() ? NEVER.pipe(startWith(undefined)) : undefined;
        }
        const isNull = observable$ == null;
        if (isNull) {
            return NEVER.pipe(startWith(observable$));
        }
        return (isObs ? observable$ : from(observable$)).pipe((o$) => {
            if (!firstRun || (withSuspense && withSuspense())) {
                return o$.pipe(startWith(undefined));
            }
            return o$;
        });
    }), tap(() => (firstRun = false)), filter((v) => v !== undefined), switchMap((o) => {
        return o.pipe(distinctUntilChanged(), rxMaterialize(), map((notification) => {
            latestNextValue =
                notification.kind === "next" /* Next */
                    ? notification.value
                    : latestNextValue;
            if (notification.kind === "next" /* Next */ &&
                latestNextValue === undefined) {
                return toRxSuspenseNotification(latestNextValue);
            }
            return Object.assign(Object.assign({}, notification), { value: latestNextValue });
        }));
    }));
    return {
        next(observable) {
            observablesSubject.next(observable);
        },
        values$,
    };
}
//# sourceMappingURL=create-template-notifier.js.map